import mrarbgrad as mag
from numpy import *
from matplotlib.pyplot import *
from numpy.linalg import norm

gamma = 42.5756e6
fov = 0.256
nPix = 256
dtGrad = 10e-6
dtADC = 2.5e-6
sLim = 50 * gamma * fov/nPix
gLim = 20e-3 * gamma * fov/nPix
# gLim = 1/nPix/dtADC
argCom = dict(fov=fov, nPix=nPix, sLim=sLim, gLim=gLim, dt=dtGrad)

mag.setSolverMtg(0) # set solver, 0 for proposed method, 1 for baseline method, baseline method may be removed due to copyright reason
mag.setTrajRev(0) # reverse the trajectory
mag.setGoldAng(1) # enable golden-angle interleaving (only for 2D)
mag.setShuf(0) # enable TR-shuffling
mag.setMaxG0(0) # use maximum possible initial gradient amplitude
mag.setMaxG1(0) # use maximum possible final gradient amplitude
mag.setMagOverSamp(8) # set oversampling ratio (default: 8, reduce only if hardware performance is too low to support real-time)
mag.setMagSFS(0) # disable DTFBS (experimental)
mag.setMagGradRep(1) # enable gradient reparameterization (experimental)
mag.setMagTrajRep(1) # enable trajectory reparameterization (experimental)
mag.setDbgPrint(1) # enable debug info (for benchmark purpose)

# calculate gradient
# lstArrK0, lstArrGrad = mag.getG_Spiral(is3D=0, **argCom); nAx = 2
# lstArrK0, lstArrGrad = mag.getG_VarDenSpiral(is3D=0, **argCom); nAx = 2
lstArrK0, lstArrGrad = mag.getG_VarDenSpiral_RT(is3D=0, **argCom); nAx = 2
# lstArrK0, lstArrGrad = mag.getG_Rosette(is3D=0, **argCom); nAx = 2
# lstArrK0, lstArrGrad = mag.getG_Rosette_Trad(**argCom); nAx = 2
# lstArrK0, lstArrGrad = mag.getG_Shell3d(**argCom); nAx = 3
# lstArrK0, lstArrGrad = mag.getG_Yarnball(**argCom); nAx = 3
# lstArrK0, lstArrGrad = mag.getG_Seiffert(**argCom); nAx = 3
# lstArrK0, lstArrGrad = mag.getG_Cones(**argCom); nAx = 3

print("")
print(f"Intlea Num.: {len(lstArrGrad)}")
nRO_Max = amax([arrG.shape[0] for arrG in lstArrGrad])
print(f"Tacq: {nRO_Max*dtGrad*1e3:.3f} ms")
tTR = (nRO_Max*dtGrad + 2e-3)
print(f"TR: {tTR*1e3:.3f} ms")
tScan = tTR*len(lstArrGrad)
print(f"Tscan: {tScan:.3e} s")

# derive shape parameter
if nAx==2:
    lstArrK0 = [arrK0[:2] for arrK0 in lstArrK0]
    lstArrGrad = [arrG[:,:2] for arrG in lstArrGrad]
nRO, nAx = lstArrGrad[0].shape

# derive slewrate
lstArrSlew = [diff(arrG, axis=0)/dtGrad for arrG in lstArrGrad]
sMax = max(norm(concatenate(lstArrSlew)/gamma*nPix/fov, axis=-1))
gMax = max(norm(concatenate(lstArrGrad)/gamma*nPix/fov, axis=-1))

# derive trajectory
lstArrK = []
for arrK0, arrGrad in zip(lstArrK0, lstArrGrad):
    arrK, _ = mag.cvtGrad2Traj(arrGrad, dtGrad, dtADC)
    arrK += arrK0
    lstArrK.append(arrK)

iArrK = len(lstArrK)*2//3

# k-space and g-space
figure(figsize=(18,9), dpi=120)

subplot(221, projection="3d" if nAx==3 else None)
plot(*lstArrK[iArrK].T, ".-")
axis("equal")
grid("on")
title(f"kspace {1}/{len(lstArrGrad)}")

subplot(223, projection="3d" if nAx==3 else None)
plot(*lstArrGrad[iArrK].T, ".-")
axis("equal")
grid("on")
title("gspace")

# gradient and slewrate
subplot(222)
for iAx in range(nAx):
    plot(lstArrGrad[iArrK][:,iAx]/gamma*nPix/fov, ".-")
grid("on")
title(f"Gradient")

subplot(224)
plot(norm(lstArrSlew[iArrK],axis=-1)/gamma*nPix/fov, ".-", c="tab:blue")
ylim(sLim/gamma*nPix/fov*0.9, sLim/gamma*nPix/fov*1.1)
grid("on")

twinx()
plot(norm(lstArrGrad[iArrK],axis=-1)/gamma*nPix/fov*1e3, ".-", c="tab:orange")
ylim(gLim/gamma*nPix/fov*0.9*1e3, gLim/gamma*nPix/fov*1.1*1e3)
grid("on")
title(f"Grad & Slew amp., max grad:{gMax*1e3:.3f}, max slew:{sMax:.3f}")

subplots_adjust(0.05,0.1,0.95,0.9, 0.2, 0.2)

show()



from numpy import *
from numpy.typing import *
from matplotlib.pyplot import *

goldrat = (1+sqrt(5))/2
goldang = (2*pi)/(1+goldrat)

getGoldrat = lambda: goldrat
getGoldang = lambda: goldang

def rand3d(i:int|NDArray, nAx:int=3, kx=sqrt(2), ky=sqrt(3), kz=sqrt(7)) -> NDArray:
    return (hstack if size(i)==1 else vstack)\
    ([
        (i**1 * 1/(1+kx))%1,
        (i**2 * 1/(1+ky))%1,
        (i**3 * 1/(1+kz))%1
    ][:nAx]).T

def cvtGrad2Traj(arrG:NDArray, dtGrad:int|float, dtADC:int|float, nShift:int|float=1.0) -> tuple[NDArray, NDArray]:
    """
    # description:
    interpolate gradient waveform and calculate trajectory

    # parameter
    `arrG`: array of gradient waveform
    `dtGrad`, `dtADC`: temporal resolution of gradient system and ADC

    # return:
    interpolated trajectory and gradient
    """
    dtShift = nShift*dtADC
    nGrad, nDim = arrG.shape
    nADC = int(dtGrad/dtADC)*(nGrad-1)
    arrG_Resamp = zeros([nADC,nDim], dtype=float64)
    for iDim in range(nDim):
        arrG_Resamp[:,iDim] = interp(dtADC*arange(nADC)+dtShift, dtGrad*arange(nGrad), arrG[:,iDim])
    arrDk = zeros_like(arrG_Resamp)
    arrDk[0,:] = (arrG[0,:] + arrG_Resamp[0,:])*dtShift/2
    arrDk[1:,:] = (arrG_Resamp[:-1] + arrG_Resamp[1:])*dtADC/2
    arrK = cumsum(arrDk,axis=0)
    return arrK, arrG_Resamp

def delGrad(arrG:NDArray, tau:int|float) -> NDArray:
    """
    # description:
    delay the input gradient waveform by time constant tau

    # parameter
    `arrG`: array of single gradient waveform
    `tau`: time constant in RL circuit transfer function

    # return:
    delayed gradient waveform
    """
    assert arrG.ndim == 2, "only single gradient waveform is supported."
    if tau == 0: return arrG.copy() # avoid divided-by-0 later
    nPt, nAx = arrG.shape

    # perform oversample to get better impluse response profile
    ov = clip(10/tau, 1, 1e3).astype(int64) # the smaller the ov, the bigger oversampling is needed
    arrG_ov = zeros([nPt*ov,nAx], dtype=arrG.dtype)
    for iAx in range(nAx):
        arrG_ov[:,iAx] = interp(linspace(0,nPt,nPt*ov,0), linspace(0,nPt,nPt,0), arrG[:,iAx]) # oversample
    nPt *= ov
    tau *= ov

    # derive impluse response of RL circuit
    arrG_Pad = zeros_like(arrG_ov)
    arrG_Pad[:nPt//2,:] = arrG_ov[-1:,:]
    arrG_Pad[nPt//2:,:] = arrG_ov[:1,:]
    arrG_ov = concatenate([arrG_ov, arrG_Pad], axis=0)
    arrT = linspace(0,2*nPt,2*nPt,0) + 0.5
    arrImpResRL = (1/tau)*exp(-arrT/tau)
    if abs(arrImpResRL.sum() - 1) > 1e-2: raise ValueError(f"arrImpResRL.sum() = {arrImpResRL.sum():.2f} (supposed to be 1) (tau too small or too large)")
    
    # perform convolution between input waveform and impulse response
    arrG_ov = fft.ifft(fft.fft(arrG_ov,axis=0)*fft.fft(arrImpResRL)[:,newaxis], axis=0).real
    
    # de-oversample
    arrG = arrG_ov[:nPt:ov,:]

    return arrG

def rotate(arr:NDArray, ang:float64, axis:int64) -> NDArray:
    if axis==0: # x
        matRot = array([
            [1, 0, 0],
            [0, cos(ang), -sin(ang)],
            [0, sin(ang), cos(ang)],
        ], dtype=float64)
    elif axis==1: # y
        matRot = array([
            [cos(ang), 0, sin(ang)],
            [0, 1, 0],
            [-sin(ang), 0, cos(ang)]
        ], dtype=float64)
    elif axis==2: # z
        matRot = array([
            [cos(ang), -sin(ang), 0],
            [sin(ang), cos(ang), 0],
            [0, 0, 1]
        ], dtype=float64)
    else:
        raise ValueError("axis should be 0, 1, 2 (denotes for x, y, z)")
    
    return arr@matRot.T

def _walsh(b:float64, k:float64, x:float64) -> float64:
    assert x>=0 and x<1
    
    # Convert k to its base-b representation
    lstKai = []
    while k > 0:
        lstKai.append(k % b)
        k //= b
    lstKai = lstKai[::-1]  # Reverse to get the correct order
    nDig = len(lstKai)
    
    # Convert x to its base-b fractional representation
    lstX = []
    for iDig in range(nDig):
        x *= b
        lstX.append(int(x))
        x -= int(x)
        
    return exp(2*pi*1j*inner(lstKai,lstX)/b)

def _calDiaphony(arrX:NDArray, b:float64=2) -> float64: # b-adic diaphony
    assert any(arrX>=0) and any(arrX<1)
    N, s = arrX.shape
    
    nume = 0
    deno = 0
    for vecK in ndindex(*([2] * s)):  # Iterate over all k in [0, 10)^s
        if all(vecK == zeros_like(vecK)): continue  # Skip k = 0

        # Compute weight r_b(k)
        r = prod([b**-floor(log(k+1)/log(b)) if k > 0 else 1 for k in vecK])

        # Compute Walsh coefficient
        meaWalsh = 0
        for vecX in arrX:
            meaWalsh += prod([_walsh(b, k, x) for k, x in zip(vecK, vecX)])
        meaWalsh /= N

        nume += r**2 * abs(meaWalsh)**2
        deno += r**2
    # deno = (1+b)**s - 1 # original implement, abandoned because it doesn't satisfy F_1 = 1

    diaphony = sqrt(nume/deno)
    return diaphony

def _calSphFibPt(nF:int64=250) -> NDArray: # get spherical Fibonacci points
    lstPtFb = []
    for iIntlea in range(nF):
        k = iIntlea - nF/2
        sf = k/(nF//2)
        cf = sqrt(((nF//2)+k)*((nF//2)-k))/(nF//2)
        phi = (1+sqrt(5))/2
        tht= 2*pi*k/phi
        
        xf = cf*sin(tht)
        yf = cf*cos(tht)
        zf = sf
        
        lstPtFb.append(array([xf,yf,zf]))
        
    return array(lstPtFb)

def _calJacElip(arrU:NDArray, m:float64) -> tuple[NDArray, NDArray]: # calculate Jacobi elliptic functions sn(u,m) and cn(u,m) numerically
    lstA = [1]
    lstB = [sqrt(1-m)]
    lstC = [0]
    while abs(lstB[-1]-lstA[-1]) > 1e-8:
        aNew = (lstA[-1]+lstB[-1])/2
        bNew = sqrt(lstA[-1]*lstB[-1])
        cNew = (lstA[-1]-lstB[-1])/2
        lstA.append(aNew)
        lstB.append(bNew)
        lstC.append(cNew)
    N = len(lstA) - 1
    lstPhi = [2**N*lstA[N]*arrU]*(N+1)
    for n in range(N,0,-1):
        lstPhi[n-1] = (1/2)*(lstPhi[n] + arcsin(lstC[n]/lstA[n]*sin(lstPhi[n])))
    arrAm = lstPhi[0]
    arrSn = sin(arrAm)
    arrCn = cos(arrAm)
    
    return arrSn, arrCn

def _calCompElipInt(m:float64) -> float64: # calculate complete Elliptical integral of the first kind
    lstA = [1]
    lstB = [sqrt(1-m)]
    while abs(lstB[-1]-lstA[-1]) > 1e-8:
        aNew = (lstA[-1]+lstB[-1])/2
        bNew = sqrt(lstA[-1]*lstB[-1])
        lstA.append(aNew)
        lstB.append(bNew)
    return pi/2/lstA[-1]





from .Function import calGrad4ExFunc, calGrad4ExSamp
from .Function import getG_Cones, getG_Rosette, getG_Rosette_Trad, getG_Seiffert, getG_Shell3d, getG_Spiral, getG_VarDenSpiral, getG_VarDenSpiral_RT, getG_Yarnball
from .Function import setSolverMtg, setTrajRev, setGoldAng, setShuf, setMaxG0, setMaxG1, setExGEnd, setMagOverSamp, setMagSFS, setMagGradRep, setMagTrajRep, setDbgPrint
from .Utility import _calDiaphony, rotate, _calJacElip, _calCompElipInt, _calSphFibPt, cvtGrad2Traj, getGoldang, getGoldrat, rand3d

from . import trajfunc


from numpy import *
from numpy.linalg import norm
from matplotlib.pyplot import *
import mrarbgrad as mag

gamma = 42.5756e6
fov = 0.256
nPix = 256
dtGrad = 10e-6
dtADC = 2.5e-6
sLim = 50 * gamma * fov/nPix
gLim = 20e-3 * gamma * fov/nPix
# gLim = 1/nPix/dtADC

# Rosette
om1 = 5*pi
om2 = 3*pi
pLim = [0,1]
def TrajFunc(t):
    rho = 0.5*sin(om1*t)
    return array\
    ([
        rho*cos(om2*t),
        rho*sin(om2*t),
        zeros_like(t)
    ])

# sample
arrP = linspace(pLim[0], pLim[1], 1000)
arrK = TrajFunc(arrP).T
nAx = 2

# derive slew-rate constrained trajectory
for i in range(1):
    arrG, _ = mag.calGrad4ExSamp(False, fov, nPix, sLim, gLim, dtGrad, arrK)
nRO, _ = arrG.shape

arrS = diff(arrG, axis=0)/dtGrad
print(f"sMax: {max(norm(arrS,axis=-1))/(42.58e6)*(nPix/fov)}")

arrK, _ = mag.cvtGrad2Traj(arrG, dtGrad, dtADC, 0.5)
arrK += TrajFunc(pLim[0])
print(f"Err: {norm(arrK[-1,:]-TrajFunc(pLim[1])):.1e}")

# derive reference trajectory
arrP_Ref = linspace(pLim[0], pLim[1], int(1e4))
arrK_Ref = TrajFunc(arrP_Ref).T

# plot
figure(figsize=(20,10), dpi=120)

subplot(221, projection=None if nAx==2 else "3d")
if nAx==2: plot(arrK_Ref[:,0], arrK_Ref[:,1], "--", label="K_Ref")
if nAx==3: plot(arrK_Ref[:,0], arrK_Ref[:,1], arrK_Ref[:,2], "--", label="K_Ref")
if nAx==2: plot(arrK[:,0], arrK[:,1], ".-", label="K_Imp")
if nAx==3: plot(arrK[:,0], arrK[:,1], arrK[:,2], ".-", label="K_Imp")
xlim(-0.5,0.5)
ylim(-0.5,0.5)
axis("equal")
grid("on")
legend()
title("k-Space")

subplot(222)
for iAx in range(nAx):
    plot(arrG[:,iAx]/(42.58e6)*(nPix/fov), ".-")
grid("on")
title("Gradient")

subplot(223, projection=None if nAx==2 else "3d")
if nAx==2: plot(arrG[:,0], arrG[:,1], ".-")
if nAx==3: plot(arrG[:,0], arrG[:,1], arrG[:,2], ".-")
axis("equal")
grid("on")
title("g-Space")

subplot(224)
plot(norm(arrS,axis=-1)/(42.58e6)*(nPix/fov), ".-")
ylim(sLim/(42.58e6)*(nPix/fov)*0.9, sLim/(42.58e6)*(nPix/fov)*1.1)
grid("on")
title(f"Slewrate, max:{max(norm(arrS,axis=-1))/(42.58e6)*(nPix/fov):.3f}")

show()



from numpy import *
from numpy.typing import *

# Spiral
prange_Spiral = lambda kRhoPhi: [0e0, 0.5/kRhoPhi]

def Spiral(phi:float64, kRhoPhi:float64=0.5/(4*pi), phi0:float64=0e0):
    rho = kRhoPhi*phi
    return array([rho*cos(phi+phi0), rho*sin(phi+phi0), zeros_like(phi)]).T

# VDS
prange_VDS = lambda kRhoPhi0, kRhoPhi1: [0e0, (log(kRhoPhi1)-log(kRhoPhi0))/(2e0*(kRhoPhi1-kRhoPhi0))]

def VDS(phi:float64, kRhoPhi0:float64=0.5/(8*pi), kRhoPhi1:float64=0.5/(2*pi), phi0:float64=0e0):
    rho = kRhoPhi0*(exp(2e0*(kRhoPhi1-kRhoPhi0)*phi)-1e0)/(2e0*(kRhoPhi1-kRhoPhi0))
    return array([rho*cos(phi+phi0), rho*sin(phi+phi0), zeros_like(phi)]).T

# Rosette
prange_Rosette = [0e0, 1e0]

def Rosette(t:float64, om1:float64=5e0*pi, om2:float64=3e0*pi, phi0:float64=0e0):
    rho = 0.5*sin(om1*t)
    return array([rho*cos(om2*t+phi0), rho*sin(om2*t+phi0), zeros_like(t)]).T

# Yarnball
prange_Yarnball = lambda kRhoPhi: [0e0, 1e0/(sqrt(8e0)*kRhoPhi)]

def Yarnball(sqrtTht:float64, kRhoPhi:float64=0.5/(2e0*pi), tht0:float64=0e0, phi0:float64=0e0):
    kPhiSqrtTht, kRhoSqrtTht = sqrt(2e0), sqrt(2e0)*kRhoPhi
    tht, rho, phi = sqrtTht**2e0*sign(sqrtTht), kRhoSqrtTht*sqrtTht, kPhiSqrtTht*sqrtTht
    return array([rho*sin(tht+tht0)*cos(phi+phi0), rho*sin(tht+tht0)*sin(phi+phi0), rho*cos(tht+tht0)]).T

# Cones
prange_Cones = lambda kRhoPhi: [0e0, 0.5/kRhoPhi]

def Cones(phi:float64, kRhoPhi:float64=0.5/(4*pi), tht0:float64=pi/3e0, phi0:float64=0e0):
    rho = kRhoPhi*phi
    return array([rho*sin(tht0)*cos(phi+phi0), rho*sin(tht0)*sin(phi+phi0), rho*cos(tht0)]).T



from setuptools import setup, Extension
import numpy
from os.path import exists
from ctypes.util import find_library

useMtg = exists("./mrarbgrad_src/ext/mtg/")
useJemalloc = find_library("jemalloc") is not None

_sources = \
[
    './mrarbgrad_src/ext/utility/global.cpp',
    './mrarbgrad_src/ext/utility/v3.cpp',
    './mrarbgrad_src/ext/mag/Mag.cpp',
    './mrarbgrad_src/ext/main.cpp',
    './mrarbgrad_src/ext/mtg/mtg_functions.cpp',
    './mrarbgrad_src/ext/mtg/spline.cpp'
]
if not useMtg:
    _sources.remove('./mrarbgrad_src/ext/mtg/mtg_functions.cpp')
    _sources.remove('./mrarbgrad_src/ext/mtg/spline.cpp')

modExt = Extension\
(
    "mrarbgrad.ext", 
    sources = _sources,
    libraries = ['jemalloc'] if useJemalloc else [],
    include_dirs = ["./mrarbgrad_src/ext/", numpy.get_include()],
    define_macros = [("USE_MTG", None)] if useMtg else None,
    language = 'c++'
)

_packages = \
[
    "mrarbgrad", 
    "mrarbgrad.trajfunc", 
    "mrarbgrad.ext", 
    "mrarbgrad.ext.traj",
    "mrarbgrad.ext.mag", 
    "mrarbgrad.ext.mtg",
    "mrarbgrad.ext.utility",
]
if not useMtg:
    _packages.remove("mrarbgrad.ext.mtg")

_package_dir = \
{
    "mrarbgrad":"./mrarbgrad_src/", 
    "mrarbgrad.trajfunc":"./mrarbgrad_src/trajfunc/",
    "mrarbgrad.ext":"./mrarbgrad_src/ext/", 
    "mrarbgrad.ext.traj":"./mrarbgrad_src/ext/traj/",
    "mrarbgrad.ext.mag":"./mrarbgrad_src/ext/mag/", 
    "mrarbgrad.ext.mtg":"./mrarbgrad_src/ext/mtg/",
    "mrarbgrad.ext.utility":"./mrarbgrad_src/ext/utility/"
}

setup\
(
    name = 'mrarbgrad',
    # install_requires = ["numpy", "matplotlib"], # pip will automatically upgrade numpy if it see this, which might corrupt the environment
    ext_modules = [modExt],
    packages = _packages,
    package_dir = _package_dir,
    include_package_data = True
)
